<?php
namespace CSPro;

use JsonStreamingParser\Listener;
use Silex\Application;
use Symfony\Component\HttpFoundation\Request;
use CSPro\CSProResponse;
use CSPro\Controllers\DictionaryController;
use JsonStreamingParser\Parser;
/**
 * This basic CasesJSON implementation of a listener simply constructs an in-memory
 * representation of the JSON document at the second level,  a single case will be kept in memory
*  rather than the whole case array for chunking. The call back function can then stash the cases
*  until a chunk is accumulated and then process for inserts
 */
class SyncCasesJsonListener implements Listener
{
	const CHUNK_SIZE = 500; //Process in chunks of 500 cases
    protected $cases;
	protected $topLevelKeyValues;
	
    protected $stack;
    protected $key;
    
    protected $app;
    protected $request;
    protected $dictName;
    protected $pdo;
    protected $dictionaryController;
    
    public $direction;
    public $currentRevisionNumber;
    protected $response;
    protected $universe;
    protected $device;
    protected $lastRevision;
	protected $idColNames;
	protected $colNames;
	protected $dict;
	protected $stopProcessing  = false;
	protected $parser = null;
	protected $initialized = false;
    // Level is required so we know how nested we are.
    protected $level;

    /**
     *
     */
    public function __construct(Application $app, $dictionaryController, Request $request, $dictName)
    {
        $this->app = $app;
        $this->dictionaryController = $dictionaryController;
        $this->request = $request;
        $this->dictName = $dictName;
        $this->pdo = $app['pdo'];
        $this->cases = [];
        $this->lastRevision = null;
        $this->idColNames = null;
        $this->colNames = null;
        $this->currentRevisionNumber =null;
        $this->direction = null;
        $this->stopProcessing = false;
        $this->response = null;
        $this->initialized = false;
        
    }
    public function setResponse(CSProResponse $response){
    	$this->response = $response;
    	$response->headers->set ( 'Content-Length', strlen ( $response->getContent () ) );
    }
    public function getResponse(){
    	return $this->response;
    }
    public function setParser(Parser $parser){
    	$this->parser = $parser;
    }
    public function stopProcessing(){
    	$this->stopProcessing = false;
    	$this->parser->stop();
    }
    public function initSyncCases(){
    	$this->initialized = true;
    	$this->app ['monolog']->addInfo ( 'initSyncCases!!!');
    	$this->universe = empty ( $this->topLevelKeyValues ['universe'] ) ? "" : $this->topLevelKeyValues ['universe'];
    	$this->direction = $this->topLevelKeyValues ['direction'];
    	$this->device = $this->topLevelKeyValues ['deviceId'];
    	
    	$this->dict = loadDictionary ( $this->app, $this->dictName );
    	
    	$this->idColNames = getCaseIdColumnNames ( $this->dict );
    	$this->colNames = array (
    			'questionnaire',
    			'caseids',
    			'label',
    			'revision',
    			'deleted',
    			'verified',
    			'partial_save_mode',
    			'partial_save_field_name',
    			'partial_save_level_key',
    			'partial_save_record_occurrence',
    			'partial_save_item_occurrence',
    			'partial_save_subitem_occurrence',
    			'clock'
    	);
    	$this->colNames = array_merge ( $this->colNames, $this->idColNames );
    	
    	// Check x-csw-if-revision-exists header to see if this an update to a previous sync
    	$this->lastRevision = null;
    	$lastSyncETag = $this->request->headers->get ('x-csw-if-revision-exists');
    	if ($lastSyncETag) {
    		// Find the previous sync - etag in the x-csw-if-revision-exists is just the revision number
    		$this->lastRevision = getSyncHistoryByRevisionNumber ( $this->pdo, $this->dictName, $lastSyncETag );
    		if (! $this->lastRevision) {
    			// Didn't find the previous sync - server and client sync history do not match
    			// Return an error, client can follow up with a full sync (no x-csw-if-revision-exists)
    			// Provide the last sync that server has record of (if there is one)
    			// so that client can possibly synced based off that revision
    			
    			//Let the parser know to stop processing
    			$lastSyncForDevice = getLastSyncForDevice ( $this->pdo,  $this->dictName, $this->device );
    			if ($lastSyncForDevice){
    				 $this->setResponse(new CSProResponse ( json_encode ( $lastSyncForDevice ), 412 ));
    				 $this->stopProcessing();
    				 return;
    			}
    			else{
    				 $this->setResponse( new CSProResponse ( json_encode ( array (), JSON_FORCE_OBJECT ) ));
    				 $this->stopProcessing();
    				 return;
    			}
    		}
    	}
    	
    	// insert a row into the sync history with the new version
    	//SynchHistoryEntry out of transaction to prevent dead locks
    	$this->currentRevisionNumber = addSyncHistoryEntry ( $this->app, $this->device, $this->dictName, $this->direction, $this->universe );
    }
    
    public function finalizeSyncCases()
    {
    	if( $this->stopProcessing == true)
    		return;
    	$this->pdo->commit();
    	$this->app ['monolog']->addInfo ( 'Done Syncing!!');
    	$this->processSyncResponse();
    }
    
    public function processSyncResponse()
    {
    	if( $this->stopProcessing == true)
    		return;
    	$response = null;
    	
    	//if only first level JSON is available because of "ClientCases":[] element is missin for "GET"
    	//initSynccases has to be run to set the top level values
    	
    	if($this->initialized == false){
    		$this->initSyncCases();
    	}
    	
    	// return the cases that were added or modified since the lastRevision
    	if ($this->direction == 'get' || $this->direction == 'both') {
    		$strWhere = ' WHERE (revision > :lastRevision AND revision < :currentRevision) ';
    		
    		
    		// If direction changed from last sync then ignore history
    		if ($this->lastRevision && $this->direction != $this->lastRevision->direction)
    			$this->lastRevision = null;
    				
    			if (! empty ( $this->universe)) {
    				$strWhere = $strWhere . ' AND (caseids LIKE :universe) ';
    				$universe = $this->universe . '%';
    				if ($this->lastRevision && ! isUniverseMoreRestrictiveOrSame ($this->universe, $this->lastRevision->universe )) {
    					$this->lastRevision = null;
    				}

    				$bind = array (
    						'lastRevision' => array (
    								'lastRevision' => ($this->lastRevision ? $this->lastRevision->revisionNumber : - 1)
    						),
    						'currentRevision' => array (
    								'currentRevision' => $this->currentRevisionNumber
    						),
    						'universe' => array (
    								'universe' => $universe
    						)
    				);
    			} else {
    				if ($this->lastRevision && ! isUniverseMoreRestrictiveOrSame ($this->universe, $this->lastRevision->universe )) {
    					$lastRevision = null;
    				}
    				$bind = array (
    						'lastRevision' => array (
    								'lastRevision' => ($this->lastRevision ? $this->lastRevision->revisionNumber : - 1)
    						),
    						'currentRevision' => array (
    								'currentRevision' => $this->currentRevisionNumber
    						)
    				);
    			}
				
    			// the statement to prepare
    			$stm = 'SELECT LCASE(CONCAT_WS("-", LEFT(HEX(guid), 8), MID(HEX(guid), 9,4), MID(HEX(guid), 13,4), MID(HEX(guid), 17,4), RIGHT(HEX(guid), 12))) as id,
										   questionnaire as data, caseids, label, deleted, verified,
										   partial_save_mode, partial_save_field_name, partial_save_level_key,
										   partial_save_record_occurrence, partial_save_item_occurrence, partial_save_subitem_occurrence,
										   clock FROM ' . $this->dictName . $strWhere . 'ORDER BY guid';
				
    			// do bind values. when universe or limit exist
    				
    			$result = $this->pdo->fetchAll ( $stm, $bind );
    				
    			// getCaseNotes from the notes table and add to each case
    			// getCaseNotes assumes caseid guids are sorted in asc - the default order by for mysql
    			$this->dictionaryController->getCaseNotes ( $this->app, $result, $this->dictName . '_notes' );
    				
    			// for each row
    			prepareResultSetForJSON ( $result );
    			$serverCases = array ();
    			$serverCases ['serverCases'] = $result;
    				
    			$response = new CSProResponse ( json_encode ( $serverCases ) );
    	} else {
    		// No server cases, send empty json object
    		$this->app ['monolog']->addInfo ( 'sending empty json');
    		$response = new CSProResponse ( json_encode ( array (), JSON_FORCE_OBJECT ) );
    	}
    	$response->headers->set ( 'ETag', $this->currentRevisionNumber );
    	$response->headers->set ( 'Content-Length', strlen ( $response->getContent () ) );
    	$this->setResponse($response);
    	
    }
    
    public function validateJSON ()
    {
    	$uri = '#/definitions/SyncRequest';
    	$syncRequest = $this->topLevelKeyValues;
    	$syncRequest['clientCases'] = $this->cases;
    	//var_dump( $this->cases);
    	//TOOD: optimize the array to object JSON  conversion
    	$syncJSONobject = json_decode(json_encode($syncRequest), false);
    	//var_dump( $syncJSONobject);
    	validateDecodedJSON ( $this->app, $syncJSONobject, $uri );
    }
    
    function processCasesInsertOrUpdate()
    {
    	if( $this->stopProcessing == true)
    		return;
    	$this->app ['monolog']->addInfo ( 'processCasesInsertOrUpdate!!!');
    	 
    	$this->validateJSON();
    	// echo 'deviceId: ' . $params['deviceId'] . ' lastRevision: ' . $lastRevision .' currentRevision: ' . $currentRevision;
    	$params ['clientCases'] = $this->cases;
    	if (($this->direction == 'put' || $this->direction == 'both') && ! empty ( $params ['clientCases'] )) {
    		// for each row get the record list array to multi-line string for the questionnaire data
    		prepareJSONForInsertOrUpdate ( $this->app, $this->dictName, $params ['clientCases'] );
    		// add the cases that do not exist using the new revision#
    		// update the cases the cases that exist using the new revision#
    		$insertQuery = array ();
    		$insertData = array ();
    		$n = 0;
    		
    		// echo 'Number of Cases: ' . count($params['clientCases']);
    		$stm = 'INSERT INTO ' . $this->dictName . ' (`guid`,' . implode ( ',', array_map ( function ($col) {
    			return "`$col`";
    		}, $this->colNames ) ) . ') VALUES ';
    		foreach ( $params ['clientCases'] as $row ) {
    			// (UNHEX(REPLACE(:guid,"-","")),:questionnaire,:revision,:deleted)
    			$insertQuery [] = '(UNHEX(REPLACE(:guid' . $n . ',"-","")),' . implode ( ',', array_map ( function ($col) use ($n) {
    				return ":$col$n";
    			}, $this->colNames ) ) . ')';
    			$insertData ['guid' . $n] = $row ['id'];
    			$insertData ['questionnaire' . $n] = $row ['data'];
    			$insertData ['caseids' . $n] = $row ['caseids'];
    			$insertData ['label' . $n] = $row ['label'];
    			$insertData ['revision' . $n] = $this->currentRevisionNumber;
    			$insertData ['deleted' . $n] = $row ['deleted'] ? 1 : 0;
    			$insertData ['verified' . $n] = $row ['verified'] ? 1 : 0;
    	
    			$insertData ['partial_save_mode' . $n] = null;
    			$insertData ['partial_save_field_name' . $n] = null;
    			$insertData ['partial_save_level_key' . $n] = null;
    			$insertData ['partial_save_record_occurrence' . $n] = null;
    			$insertData ['partial_save_item_occurrence' . $n] = null;
    			$insertData ['partial_save_subitem_occurrence' . $n] = null;
    	
    			if (isset ( $row ['partialSave'] )) {
    				$insertData ['partial_save_mode' . $n] = $row ['partialSave'] ['mode'];
    				$insertData ['partial_save_field_name' . $n] = $row ['partialSave'] ['field'] ['name'];
    				$insertData ['partial_save_level_key' . $n] = isset ( $row ['partialSave'] ['field'] ['levelKey'] ) ? $row ['partialSave'] ['field'] ['levelKey'] : "";
    				$insertData ['partial_save_record_occurrence' . $n] = isset ( $row ['partialSave'] ['field'] ['recordOccurrence'] ) ? $row ['partialSave'] ['field'] ['recordOccurrence'] : 0;
    				$insertData ['partial_save_item_occurrence' . $n] = isset ( $row ['partialSave'] ['field'] ['itemOccurrence'] ) ? $row ['partialSave'] ['field'] ['itemOccurrence'] : 0;
    				$insertData ['partial_save_subitem_occurrence' . $n] = isset ( $row ['partialSave'] ['field'] ['subitemOccurrence'] ) ? $row ['partialSave'] ['field'] ['subitemOccurrence'] : 0;
    			}
    	
    			$insertData ['clock' . $n] = $row ['clock'];
    			$idColVals = getCaseIdColumnValues ( $this->dict, $row ['data'] );
    			for($iIdCol = 0; $iIdCol < count ( $this->idColNames ); ++ $iIdCol) {
    				$insertData [$this->idColNames [$iIdCol] . $n] = $idColVals [$iIdCol];
    			}
    			$n ++;
    		}
    		if (! empty ( $insertQuery )) {
    			$stm .= implode ( ', ', $insertQuery );
    	
    			$stm .= ' ON DUPLICATE KEY UPDATE `guid`=VALUES(`guid`), ';
    			$stm .= implode ( ',', array_map ( function ($col) {
    				return "`$col`=VALUES(`$col`)";
    			}, $this->colNames ) );
    				$stm .= ';';
    	
    				// $app['pdo']->getProfiler()->setActive(true); //Uncomment this line to activate profiler
    				// prepare insert statement and bind values
    				$stmt = $this->pdo->prepare ( $stm );
    				$result = $stmt->execute ( $insertData ); // true if successful
    				 
    				// add notes to the notes table
    				$this->dictionaryController->addCaseNotes ( $this->app, $params ['clientCases'], $this->dictName . '_notes' );
    				// dumpProfilerOutput($app) //Uncomment this line to dump the profiler results
    		}
    	}
    }
   

    public function startDocument()
    {
    	$this->app['monolog']->addInfo ( 'startDocument!!!');
    	 
        $this->stack = [];
        $this->level = 0;
        // Key is an array so that we can can remember keys per level to avoid
        // it being reset when processing child keys.
        $this->key = [];
    }

    public function endDocument()
    {
    	if( $this->stopProcessing == true)
    		return;
    	$this->app['monolog']->addInfo ( 'endDocument!!!');
        //process the final chunk
    	if(count($this->cases) != 0){
    		$this->processCasesInsertOrUpdate();
    		$this->cases = array(); //clear the array after processing the chunk
    	}
        // call end sync
         $this->finalizeSyncCases();
    }

    public function startObject()
    {
		if ($this->level == 1 && empty($this->topLevelKeyValues)) {//if about to start second level objects processing store the topLevelKeys before reset.
			$obj = array_pop($this->stack);
			$this->topLevelKeyValues = $obj;
			$this->stack = $obj; 
			$this->initSyncCases();
		}
        $this->level++;
        $this->stack[] = [];
        // Reset the stack when entering the second level
        if ($this->level == 2) {
            $this->stack = [];
            $this->key[$this->level] = null;
        }
    }

    public function endObject()
    {
        $this->level--;
        $obj = array_pop($this->stack);
        if (empty($this->stack)) {
            // doc is DONE!
            if($this->level == 2) //store cases
            	$this->cases[] = $obj;
            else if($this->level == 0 && empty($this->topLevelKeyValues)){
            	//if the level 2 "clientcases" element is missing . Level 1 elements neeed to 
            	//process for "get". So, set the toplevel keys 
            	$this->topLevelKeyValues = $obj;
            }
        } else {
            $this->value($obj);
        }
        // strore the cases for processing as chunks
        if ($this->level == 2) {
           if(count($this->cases) >= self::CHUNK_SIZE){
           	$this->processCasesInsertOrUpdate();
           	$this->cases = array(); //clear the array after processing the chunk
           }
           	
        }
    }

    public function startArray()
    {
        $this->startObject();
    }

    public function endArray()
    {
        $this->endObject();
    }

    /**
     * @param string $key
     */
    public function key($key)
    {
        $this->key[$this->level] = $key;
    }

    /**
     * Value may be a string, integer, boolean, null
     * @param mixed $value
     */
    public function value($value)
    {
        $obj = array_pop($this->stack);
        if (!empty($this->key[$this->level])) {
            $obj[$this->key[$this->level]] = $value;
            $this->key[$this->level] = null;
        } else {
            $obj[] = $value;
        }
        $this->stack[] = $obj;
    }

    public function whitespace($whitespace)
    {
        // do nothing
    }
}
